;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "deflate" "0.0.5"
  "The DEFLATE compression algorithm in pure Emacs LISP."
  '((dash  "2.0.0")
    (emacs "25.1"))
  :url "https://github.com/skuro/deflate"
  :commit "4896cdf0c1d031404c6705f52c03f048444ff927"
  :revdesc "0.0.5-0-g4896cdf0c1d0"
  :keywords '("files" "tools")
  :authors '(("Carlo Sciolla" . "carlo.sciolla@gmail.com"))
  :maintainers '(("Carlo Sciolla" . "carlo.sciolla@gmail.com")))
